#!/usr/bin/env perl
# Copyright (C) 2011-2012 by Tomasz Konojacki
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Setup stuff...
package Kitsune;

use strict;
use warnings;
use utf8;
use open qw(:std :utf8);

our $VERSION = '1.1.3';

################################################################################
# KITSUNE CONFIGURATION                                                        #
################################################################################

# Path to directory where entries are stored:
# (Note that under Windows and IIS you need to make sure that folder containing
#  kitsune.pl [or whatever you name this file] is converted to application)
# Examples:
#   '/home/xenu/entries/'
#   'entries'
#   '.'

my $entries_path = 'entries';

# CSS URL:
# Examples:
#   'http://foobar.com/foo.css'
#   'whatever.css'
#   ''

my $css_url = '';

# Entry extension:
# Examples:
#   'txt'
#   'dat'

my $entry_extension = 'txt';

# Name of blog:
# Examples:
#   'Peter's blog'

my $blog_name = 'Kitsune powered blog';

# Description of blog (for RSS feed):
# Examples:
#   'Blog about me and my friends'

my $blog_description = 'Kitsune powered blog';

# Footer:
# Examples:
#   'Copyright © Someone'

my $footer = 'Copyright © Someone';

# Language of blog:
# Examples:
#   'en-us'
#   'en-gb'

my $blog_language = 'en-us';

# Number of entries on one page:
# Examples:
#   10
#   20

my $max_entries = 10;

# Number of entries in RSS:
# Examples:
#   10
#   20

my $rss_entries = 10;


################################################################################
# END OF CONFIGURATION, DON'T EDIT ANYTHING BELOW THAT LINE!                   #
################################################################################

# Load neccssary modules.
use CGI qw(-utf8 -no_xhtml);
use CGI::Carp qw(fatalsToBrowser);
use CGI::Util qw(escape unescape);
use Encode qw(encode decode);
use File::stat;
use IO::Dir;
use IO::File;
use Time::Piece;

sub load_entries {
    # This subroutine loads entries to array.
    my($first_entry, $last_entry, $entries_ref) = @_;
    my(@entries, $dir, $number_of_entries);
    # Open directory where entries are contained.
    $dir = IO::Dir->new($entries_path);
    if (!defined $dir) { die "Unable to open directory \"$entries_path\"" }
    while (defined($_ = $dir->read)) {
        # We don't need directories and non-entry files.
        next if -d "$entries_path/$_";
        next unless m/\.$entry_extension$/;
        # Silently ignore unreadable files.
        next unless -r "$entries_path/$_";
        # Load filenames and creation dates into entries array.
        $entries[++$#entries]{filename} = $entries_path . '/' . $_;
        $entries[$#entries]{created} = localtime(stat("$entries_path/$_")->mtime);
        s/\.$entry_extension$//;
        $entries[$#entries]{name} = $_;
    }
    $dir->close;
    # Sort entries array by entry creation date.
    @entries = sort {$b->{created} <=> $a->{created}} @entries;
    # Check is there enough entries for requested extent, if not make
    # $last_entry smaller.
    if ($#entries >= $last_entry) {
        1;
    }
    elsif (($#entries < $last_entry) && ($first_entry <= $#entries)) {
        $last_entry = $#entries;
    }
    else {
        return 0;
    }
    # Expected number of entries (will be returned by this subroutine):
    $number_of_entries = $last_entry - $first_entry + 1;
    # Load requested entries into array.
    for ($first_entry..$last_entry) {
        my ($filename, $fh);
        $filename = $entries[$_]{filename};
        die "Unable to open \"$filename\"" unless -r $filename;
        $fh = IO::File->new($filename, '<:utf8') or die $!;
        # Read entry title.
        $entries[$_]{title} = <$fh>;
        {
            # Load rest of file.
            local $/;
            $entries[$_]{text} = <$fh>;
            $fh->close;
        }
        # It prevents script from using undefined value if file is empty:
        $entries[$_]{title} ||= '';
        $entries[$_]{text} ||= '';
    }
    @{$entries_ref} = @entries;
    return $number_of_entries;
}

sub show_rss_feed {
    # This subroutine displays RSS feed.
    my $cgi = shift;
    my(@entries, $last_entry, $number_of_entries, $self_url);
    $self_url = $cgi->url;
    $last_entry = $rss_entries - 1;
    $number_of_entries = load_entries(0, $last_entry, \@entries);
    print $cgi->header(
            -charset => 'UTF-8',
            -type    => 'application/rss+xml'
        ),
        '<?xml version="1.0"?>',
        '<!DOCTYPE rss PUBLIC "-//XENU//DTD RSS 0.99//EN" "http://puszcza.gnu.org.ua/software/rss099/rss-0.99.dtd">',
        '<rss version="0.99">',
        '<channel>',
        "<title>$blog_name</title>",
        "<link>$self_url</link>",
        "<description>$blog_description</description>",
        "<language>$blog_language</language>";
        for (0..$number_of_entries - 1) {
            my ($title, $name, $link, $text);
            $title = CGI::escapeHTML($entries[$_]{title});
            $name = escape($entries[$_]{name});
            $link = "$self_url?$name";
            $text = CGI::escapeHTML($entries[$_]{text});
            print '<item>',
                "<title>$title</title>",
                "<link>$link</link>",
                "<description>$text</description>",
                '</item>';
        }
        print '</channel></rss>';


}

sub show_page {
    # This subroutine displays requested page.
    my($cgi, $page) = @_;
    my(@entries, $first_entry, $last_entry, $number_of_entries, $self_url);
    # Save self URL.
    $self_url = $cgi->url(-relative => 1);
    # Calculate extent required by load_entries().
    $first_entry = $page * $max_entries;
    $last_entry = $page * $max_entries + $max_entries - 1;
    # Load entries.
    $number_of_entries = load_entries($first_entry, $last_entry, \@entries);
    die 'No entries found' unless $number_of_entries;
    # Output entries to browser.
    print $cgi->header(-charset => 'UTF-8');
    print $cgi->start_html(
            -lang => $blog_language,
            -head => $cgi->Link({
                'rel'   => 'alternate',
                'type'  => 'application/rss+xml',
                'title' => 'RSS',
                'href'  => "$self_url?rss=1"
            }),
            -title => $blog_name,
            # If $css_url is defined use stylesheet.
            $css_url ? (-style => { -src => $css_url }) : ''
        ),
        $cgi->center(
            $cgi->h1($blog_name)
        );
    for (1..$number_of_entries) {
        print $cgi->h3($entries[$first_entry]{created}),
            $cgi->h2($entries[$first_entry]{title}),
            $entries[$first_entry]{text},
            $cgi->p(
                $cgi->a({href => $self_url . '?' . escape($entries[$first_entry]{name})}, 'Permanent link to this entry')
            );
            # Script uses $first_entry as entry counter.
            $first_entry++
    }
    print $cgi->p(
            # Display link to previous page if user is not on first page.
            ($page > 0) ? ($cgi->a({href => $self_url . '?page=' . ($page - 1)}, '&lt;&lt;Previous page'), '&nbsp;&nbsp') : '',
            # Display next page only when it exists.
            (scalar @entries > ($page + 1) * $max_entries) ? $cgi->a({ href => $self_url . '?page=' . ($page + 1)}, 'Next page&gt;&gt;') : '',
        ),
        $cgi->p($footer),
        $cgi->end_html;

}

sub show_entry {
    # This subroutine displays requested entry.
    my($cgi, $entry) = @_;
    my($entry_file, $entry_creation_date, $fh, $entry_title, $entry_text, $self_url);
    # Save self URL.
    $self_url = $cgi->url(-relative => 1);
    # Generate path to requested entry, note that '/' is universal and works
    # both under Windows and UNIX.
    $entry_file = "$entries_path/$entry.$entry_extension";
    # Check if file exist and is readable.
    die "Entry \"$entry\" does not exist" unless -r $entry_file;
    # Get file cration date and use it as blog entry creation date.
    $entry_creation_date = localtime(stat($entry_file)->mtime);
    $entry_creation_date = $entry_creation_date->cdate;
    $fh = IO::File->new($entry_file, '<:utf8') or die $!;
    # Read entry title.
    $entry_title = <$fh>;
    {
        # Read rest of entry.
        local $/;
        $entry_text = <$fh>;
        $fh->close;
    }
    # It prevents script from using undefined value if file is empty:
    $entry_title ||= '';
    $entry_text ||= '';
    # Output entry to browser.
    print $cgi->header(-charset => 'UTF-8');
    print $cgi->start_html(
            -lang => $blog_language,
            -head => $cgi->Link({
                'rel'   => 'alternate',
                'type'  => 'application/rss+xml',
                'title' => 'RSS',
                'href'  => "$self_url?rss=1"
            }),
            # If $css_url is defined use stylesheet.
            -title => "$blog_name - $entry_title",
            $css_url ? (-style => { -src => $css_url }) : ''
        ),
        $cgi->center($cgi->h1($blog_name)),
        $cgi->h3($entry_creation_date),
        $cgi->h2($entry_title),
        $entry_text,
        $cgi->p($footer),
        $cgi->end_html;
}

sub run {
    # This subroutine is equivalent to C main() function.
    # Create CGI object.
    my $cgi = CGI->new;
    # If 'rss' is set to true value show RSS feed.
    if ((defined $cgi->param('rss')) && $cgi->param('rss')) {
        show_rss_feed($cgi);
    }
    # If 'page' is defined show requested page.
    elsif (defined $cgi->param('page')) {
        show_page($cgi, $cgi->param('page'));
    }
    # QUERY_STRING is defined so show requested entry.
    elsif (defined $ENV{QUERY_STRING} && ($ENV{QUERY_STRING} ne '')) {
        show_entry($cgi, decode('UTF-8', unescape($ENV{QUERY_STRING})));
    }
    # By default show first page.
    else {
        show_page($cgi, 0);
    }
}

run();